@@ -1,9 +1,9 @@
 -- | Core types and definitions
 {-# LANGUAGE GeneralizedNewtypeDeriving, FlexibleContexts,
-             ExistentialQuantification, RankNTypes, DeriveDataTypeable #-}
+             ExistentialQuantification, RankNTypes, DeriveDataTypeable,
+             DeriveGeneric #-}
 module Test.Tasty.Core where
 
-import Control.Applicative
 import Control.Exception
 import Test.Tasty.Options
 import Test.Tasty.Patterns
@@ -12,14 +12,35 @@ import Data.Monoid
 import Data.Typeable
 import qualified Data.Map as Map
 import Data.Tagged
+import GHC.Generics
 import Text.Printf
 
+-- | If a test failed, 'FailureReason' describes why
+data FailureReason
+  = TestFailed
+    -- ^ test provider indicated failure
+  | TestThrewException SomeException
+    -- ^ test resulted in an exception. Note that some test providers may
+    -- catch exceptions in order to provide more meaningful errors. In that
+    -- case, the 'FailureReason' will be 'TestFailed', not
+    -- 'TestThrewException'.
+  | TestTimedOut Integer
+    -- ^ test didn't complete in allotted time
+  deriving Show
+
+-- | Outcome of a test run
+--
+-- Note: this is isomorphic to @'Maybe' 'FailureReason'@. You can use the
+-- @generic-maybe@ package to exploit that.
+data Outcome
+  = Success -- ^ test succeeded
+  | Failure FailureReason -- ^ test failed because of the 'FailureReason'
+  deriving (Show, Generic)
+
 -- | A test result
 data Result = Result
-  { resultSuccessful :: Bool
-    -- ^
-    -- 'resultSuccessful' should be 'True' for a passed test and 'False' for
-    -- a failed one.
+  { resultOutcome :: Outcome
+    -- ^ Did the test fail? If so, why?
   , resultDescription :: String
     -- ^
     -- 'resultDescription' may contain some details about the test. For
@@ -31,6 +52,20 @@ data Result = Result
     -- information about the failure.
   }
 
+-- | 'True' for a passed test, 'False' for a failed one.
+resultSuccessful :: Result -> Bool
+resultSuccessful r =
+  case resultOutcome r of
+    Success -> True
+    Failure {} -> False
+
+-- | Shortcut for creating a 'Result' that indicates exception
+exceptionResult :: SomeException -> Result
+exceptionResult e = Result
+  { resultOutcome = Failure $ TestThrewException e
+  , resultDescription = "Exception: " ++ show e
+  }
+
 -- | Test progress information.
 --
 -- This may be used by a runner to provide some feedback to the user while
@@ -66,17 +101,21 @@ type TestName = String
 -- and how to release it (the second field).
 data ResourceSpec a = ResourceSpec (IO a) (a -> IO ())
 
+-- | A resources-related exception
 data ResourceError
   = NotRunningTests
-  | UnexpectedState String
+  | UnexpectedState String String
+  | UseOutsideOfTest
   deriving Typeable
 
 instance Show ResourceError where
   show NotRunningTests =
     "Unhandled resource. Probably a bug in the runner you're using."
-  show (UnexpectedState state) =
-    printf "Unexpected state of the resource (%s). Report as a tasty bug."
-      state
+  show (UnexpectedState where_ what) =
+    printf "Unexpected state of the resource (%s) in %s. Report as a tasty bug."
+      what where_
+  show UseOutsideOfTest =
+    "It looks like you're attempting to use a resource outside of its test. Don't do that!"
 
 instance Exception ResourceError
 
@@ -181,12 +220,6 @@ foldTestTree (TreeFold fTest fGroup fResource) opts tree =
         WithResource res tree -> fResource res $ \res -> go pat path opts (tree res)
         AskOptions f -> go pat path opts (f opts)
 
--- | Useful wrapper for use with foldTestTree
-newtype AppMonoid f = AppMonoid { getApp :: f () }
-instance Applicative f => Monoid (AppMonoid f) where
-  mempty = AppMonoid $ pure ()
-  AppMonoid f1 `mappend` AppMonoid f2 = AppMonoid $ f1 *> f2
-
 -- | Get the list of options that are relevant for a given test tree
 treeOptions :: TestTree -> [OptionDescription]
 treeOptions =
