@@ -6,14 +6,18 @@ module Test.Tasty.Runners
   , foldTestTree
   , TreeFold(..)
   , trivialFold
-  , AppMonoid(..)
   , ResourceSpec(..)
+  , module Test.Tasty.Runners.Reducers
     -- * Ingredients
   , Ingredient(..)
   , tryIngredients
   , ingredientOptions
   , ingredientsOptions
     -- * Standard console ingredients
+    -- | NOTE: the exports in this section are deprecated and will be
+    -- removed in the future. Please import "Test.Tasty.Ingredients.Basic"
+    -- if you need them.
+
     -- ** Console test reporter
   , consoleTestReporter
     -- ** Tests list
@@ -27,6 +31,9 @@ module Test.Tasty.Runners
     -- * Running tests
   , Status(..)
   , Result(..)
+  , Outcome(..)
+  , FailureReason(..)
+  , resultSuccessful
   , Progress(..)
   , StatusMap
   , launchTestTree
@@ -42,8 +49,8 @@ module Test.Tasty.Runners
 import Test.Tasty.Core
 import Test.Tasty.Run
 import Test.Tasty.Ingredients
-import Test.Tasty.CoreOptions
+import Test.Tasty.Options.Core
 import Test.Tasty.Patterns
 import Test.Tasty.CmdLine
-import Test.Tasty.Ingredients.ConsoleReporter
-import Test.Tasty.Ingredients.ListTests
+import Test.Tasty.Ingredients.Basic
+import Test.Tasty.Runners.Reducers
